#
#   Created By Ryan McDermott
#   Created on 5/4/2022
#

from indxdatalaketools import Helpers

class Wrapper:
    '''
        Wrapper for Google Cloud Storage python SDK
    '''
    __google_cloud_storage_client = None

    def __init__(self, google_clients):
        self.__google_cloud_storage_client = google_clients.cloud_storage_client

    def find_blobs_with_metadata(self, bucket, key, value, path_prefix=''):
        '''
            Finds every file with with the metadata matching the key and value pairs
            provided. Can also set a prefix to narrow down the available searches
            Args:
                key (string): The metadata key
                value (string): the metadata value
                path_prefix (string): blob path prefix
            Returns:
                list: list containing all blobs with mathcing metadata
        '''
        blobs = []
        for blob in self.__google_cloud_storage_client.list_blobs(bucket, prefix=path_prefix):
            if blob.metadata is not None and key in blob.metadata and blob.metadata[key] == value:
                blobs.append(blob.name)

        return blobs

    def check_if_bucket_exists(self, bucket_name):
        '''
            checks if the bucket name passed exists
            Args:
                bucket_name (string): the name of the bucket
            Returns:
                boolean: True if the bucket exists, False if otherwise
        '''
        try:
            self.__google_cloud_storage_client.get_bucket(bucket_name)
            return True
        except Exception as e:
            print("bucket " + bucket_name + " does not exist")
            return False

    def upload_local_file_to_gcs_bucket(self, mrn, file_path, prefix, client_uuid):
        '''
            uploads file to specified folder in data lake - prefix determines the folder
            args:
                mrn: str - user input mrn
                file_path: str - local file path of file to upload
                prefix: str - folder on data lake where we want file to be placed
            return:
                upload_result: boolean - true if file upload successful, false if failure encountered
        '''
        print(client_uuid)
        file_path           = file_path.replace("\\", "/")
        file_name           = file_path.split("/")[-1]
        blob_name           = prefix + "/" + mrn + "_" + file_name
        bucket              = self.__google_cloud_storage_client.bucket(client_uuid)
        blob                = bucket.blob(blob_name)
        upload_result       = self.__upload_blob_from_local(blob, file_path)
        if not upload_result:
            return upload_result
        return blob

    def __upload_blob_from_local(self, blob, file_path):
        '''
            applies gcs api to upload a file from local to blob on gcs
            args:
                blob: gcs blob object
                file_path: local file path of file to be uploaded
            return:
                boolean: true is success and false if failure encountered
        '''
        try:
            blob.upload_from_filename(file_path)
            return True
        except Exception as e:
            Helpers.print_error("Could not connect to GCS client " + str(e))
            return False




#
#   Created by Ryan McDermott
#   Created on 7/8/2022
#

from indxdatalaketools.PropertiesDBApi.PropertiesApi import Client

class Properties(object):
    properties_api = None
    modalities = []
    race_codes = []
    _instance = None
    

    def __init__(self):
        '''
            Should not be able to make a new instance of this singleton
        '''
        raise RuntimeError('Call instance() instead')

    @classmethod
    def instance(cls,google_clients=None):
        '''
        singleton design pattern, call this function to get the singleton
        Args:
            cls (self@Properties): class of type self Properties
            google_clients (googleClients): The authentication to every google glient service
        Returns
            Properties: The properties singleton
        '''
        if cls._instance is None:
            cls._instance = cls.__new__(cls)
            cls._instance.properties_api = Client(google_clients)
            cls._instance.modalities = cls._instance.properties_api.get_modalities()
            cls._instance.race_codes = cls._instance.properties_api.get_race_codes()
            # Put any initialization here.
        return cls._instance


    def refresh_properties(self):
        '''
            Function that makes an api call to the properties DB API to refresh the property values
            found in the modalities and race codes.
        '''
        self.modalities = self.properties_api.get_modalities()
        self.race_codes = self.properties_api.get_race_codes()


#
#   Created By Ryan McDermott
#   Created on 2/22/2022
#

from indxdatalaketools import Helpers
from indxdatalaketools.ArgumentValidation import ValidationHelpers
from indxdatalaketools.PropertiesDBApi import Properties

HL7_SEX             = ["F", "M", "O", "U"]
HL7_FYR_SEX         = ["female", "male", "other", "unknown"]


class Validator():
    '''
        Class that helps validate the fields in the patient data passed by the client
    '''

    client_uuid                         =  ''
    patients_command                    =  ''
    arguments                           = {}

    __patient_data_validation_functions = None

    def __init__(self):
        '''
            initializes the Patient Data Validator
        '''
        self.__set_up_client_id_validation_functions()


    def __set_up_client_id_validation_functions(self):
        '''
            Sets up the mapping from the arguments passed in patient_data 
            to the function that will validate the given field
            Args:
                None
            Returns:
                None
        '''
        self.__patient_data_validation_functions = {
            'CLIENT_ID': self.__validate_client,
            'MRN': self.__validate_mrn,
            'PATIENT_ID': self.__validate_patient_id,
            'FIRSTNAME': self.__validate_name,
            'MIDDLENAME': self.__validate_name,
            'LASTNAME': self.__validate_name,
            'DOB': ValidationHelpers.validate_date,
            'SEX': self.__validate_sex,
            'RACE': self.__validate_race
        }


    def validate(self, patient_data, arguments):
        '''
            Validates the patient data
            Args:
                patient_data: structure or file containing one or more patients
                    data
            Returns:
                boolean: True if all Patients have valid data, False if otherwise
        '''
        self.client_uuid        = arguments['client_id']
        self.arguments          = arguments

        json_structure = Helpers.determine_json_structure(patient_data)

        if isinstance(json_structure, list):
            return self.__validate_multiple_patients_data_json(json_structure)

        return self.__validate_patient_data_json(json_structure)
    

    def __validate_multiple_patients_data_json(self, json_structure):
        '''
            Verifies multiple patients data passed by the client
            Args:
                json_structure (list[dict]): list of patient data
            returns:
                boolean: True if all patients' data is valid
        '''
        for i in json_structure:

            if not self.__validate_patient_data_json(i):
                return False
            
        return True


    def __validate_patient_data_json(self, patient_data):
        '''
            Verifies that the json passed contains valid attributes
            Args:
                None
            Returns:
                boolean: True if the patient data is valid, False if otherwise
        '''
        if 'MRN' not in patient_data or 'CLIENT_ID' not in patient_data:
            Helpers.print_error("required fields MRN or CLIENT_ID is missing")
            return False

        for field in patient_data:
            if not self.__validate_field_in_patient_data(field, patient_data):
                return False

        return True
        

    def __validate_field_in_patient_data(self, field, patient_data):
        '''
            Validates a given fields in the patient data json structure
            Args:
                field           (string): The field we are going to validate
                patient_data    (dict): The dictionary of the patient data
            Returns:
                boolean: True if the field value is valid, False if otherwise
        '''
        # special case for PATIENT_ID because it requires multiple fields to check
        # its validity

        if field == 'PATIENT_ID':
            return self.__validate_patient_id(patient_data)

        return self.__patient_data_validation_functions[field](patient_data[field])

    
    def __validate_client(self, client):
        '''
            Validates that the Client id passed is the same as the client_id in the 
            json object
            Args:
                None
            Returns:
                boolean: True if the bucket exists, False if otherwise
        '''
        if client != self.client_uuid:
            Helpers.print_error("Invalid CLIENT_ID " + client + ": Bad Request")
            return False

        return True


    def __validate_mrn(self, mrn):
        '''
            Validates that the MRN only contains digits and letters
            Args:
                None
            Returns:
                boolean: True if the MRN is valid, False if otherwise
        '''
        if all(character.isalpha() or character.isdigit() or \
            character == '-' for character in mrn):
            return True
        
        Helpers.print_error(mrn + " is not a valid MRN: Does Not contain only letters, digits and hyphens")
        return False


    def __validate_patient_id(self, patient_data):
        '''
            Validates that the hash of the mrn and the client ID is the patient ID that was 
            passed
            Args:
                patient_data (dict): The dictionary containing all of the patients information
            Returns:
                boolean: True if the patient_id is valid, False if otherwise
        '''
        expected_patient_id = Helpers.patient_hash(patient_data["CLIENT_ID"], 
            [patient_data["MRN"]])[0]

        if expected_patient_id != patient_data["PATIENT_ID"]:
            Helpers.print_error('the Patient ID passed is not valid: Bad Request')
            return False
        
        return True


    def __validate_name(self, name):
        '''
            Validates a first, last, or middle name of a patient. This includes checking for only 
            letters and spaces
            Args:
                name (string): the name that we wish to pass
            Returns:
                boolean: True if the name is valid, False if otherwise
        '''
        if all(character.isalpha() or character.isspace() for character in name):
            return True
        
        Helpers.print_error(name + " is not valid: Does Not contain only letters and spaces")
        return False


    def __validate_race(self, race):
        '''
            function that checks the race of a patient. Race should be identified 
            using the appropriate HL7 code represented as a string.  
            e.g. 1014-0 (Lipan Apache) should be represented as the string literal 1014-0. 
            Args:
                race (string): The HL7 code for race
            Returns:
                boolean: True if race is a valid HL7 code and false if other wise
        '''
        if race in Properties.Properties.instance().race_codes:
            return True
        
        Helpers.print_error("Race " + race + " is not a HL7 race code")
        return False


    def __validate_sex(self, sex):
        '''
            function that checks the sex of a patient. Sex is defined as 
            the administrative gender using the appropriate HL7 code represented as a string.  
            F for female, M for male, and UN for undifferentiated. 
            Args:
                sex (string): The HL7 code for sex
            Returns:
                boolean: True if race is a valid HL7 code and false if other wise
        '''
        if sex in HL7_SEX or sex in HL7_FYR_SEX:
            return True
        
        Helpers.print_error("sex " + sex + " is not a HL7 sex code")
        return False



#
#   Created By Ryan McDermott
#   Created on 2/21/2022
#
#   Helper functions used in the datalake tools

import ast
import hashlib
import sys

import requests



#
#   REVIEWERS: Is there a better way to handle these cases? (missing underscore)
#
def convert_keys_in_dictionary(patient_data):
    '''
        This function converts the keys in the patient data dictionary
            To the form recognized by the API
        Args:
            - patient_data (dict): the patient data dictionary
        Returns:
            - dict: The converted dictionary
    '''
    converted_patient_data = {}

    for key, value in patient_data.items():
        new_key = key.lower().replace("_","-")
        if "name" in new_key:
            new_key = new_key.replace("name", "-name")
        if value is None:
            continue
        converted_patient_data[new_key] = value

    return converted_patient_data


def delete_blob(blob):
    '''
        connects to GCS to delete blob passed
        Args:
            blob (google.cloud.storage.blob): The blob we wish to delete
        Returns:
            boolean: True if the blob was deleted, False if otherwise
    '''
    try:
        blob.delete()
        return True
    except Exception as e:
        print_error("Could not delete blob " + str(e))
        return False


def patch_bucket(bucket):
    '''
        Try and Catch block to determine if the bucket patch was successful
        Args: 
            bucket (google.cloud.storage.Bucket): The bucket we wish to patch
        Returns:
            boolean: True if the operation was sucesfull, False if otherwise
    '''
    try:
        bucket.patch()
        return True
    except Exception as exception:
        print_error("Bucket was unable to get patched: " + str(exception))
        return False


def patch_blob(blob):
    '''
        Try and Catch block to determine if the blob patch was successful
        Args: 
            bucket (google.cloud.storage.Blob): The blob we wish to patch
        Returns:
            boolean: True if the operation was sucesfull, False if otherwise
    '''
    try:
        blob.patch()
        return True
    except Exception as exception:
        print_error("Bucket was unable to get patched: " + str(exception))
        return False


def try_bucket_exists(bucket):
    '''
        Try catch block of checking if a GCS Bucket exists
    '''
    try:
        return bucket.exists()
    except Exception as e:
        print_error("Could not check if bucket exists: " + str(e))
        return False


def try_blob_exists(blob):
    '''
        Try catch block of checking if a GCS Blob exists
    '''
    try:
        return blob.exists()
    except Exception as e:
        print_error("Could not check if blob exists: " + str(e))
        return False


def print_error(*args, **kwargs):
    '''
        Called the same was as print, this instead prints to std err
    '''
    print(*args, file=sys.stderr, **kwargs)



def create_service_account_name(client_uuid):
    '''
        Function that creates an md5 hash of the client uuid then removes the last 8 letters of it
        and adds client to the front. this will be the name for the client service account
        Args:
            client_uuid (string): the uuid of the client
        Returns:
            string: the service account name
    '''
    md5_hash                = hashlib.md5()
    encoded                 = client_uuid.encode()
    md5_hash.update(encoded)
    md_5_clientuuid         = md5_hash.hexdigest()
    service_account_name    = 'client' + str(md_5_clientuuid)[0:24]

    return service_account_name
    

def hash_contents_to_file_name(file_path, contents):
    """
        hashes the contents of the file to be used for rename the file
        Args:
            file_path (string): the path to the file, used to get extension
            contents (bytes): byte representation of the file contents
        Returns:
            string: the new file name
    """
    # for windows
    file_path = file_path.replace("\\", "/")
    extension = file_path.split("/")[-1].split(".")[-1]
    encoded_contents = contents
    hash_contents = hashlib.sha1(encoded_contents)
    name = hash_contents.hexdigest()

    return name + "." + extension


def patient_hash(client_id, patient_mrn_list):
    """
        This function takes the patients MRN and the client id's 
        UUID and creates a sha1 hash for the patient ID
        Args:
            client_id (string): The clients UUID
            patient_mrn_list (list of string): a list of Patient MRN 
        Returns:
            list: The list of strings containing all of the patient Id's 
    """
    patient_hash_list = []
    for patient in patient_mrn_list:
        for i in range(len(patient)):
            if patient[i] != '0':
                patient = patient[i:]
                break
        if patient == 'UNKNOWN':
            patient_hash_list.append('UNKNOWN')
        client_patient = client_id.strip()+patient.strip()
        encoded_patient = client_patient.encode()
        hash_patient = hashlib.sha1(encoded_patient)
        patient_id = hash_patient.hexdigest()
        patient_hash_list.append(patient_id)
    
    return patient_hash_list


def determine_json_structure(json_data):
    """
        determines if meta data is a string object or a filepath
        Args:
            metadata (string): either file path, json string of metadata, or dictionary
        Returns:
            dict: dictionary depiction of metadata
    """
    if isinstance(json_data, dict) or isinstance(json_data, list):
        return json_data

    if "{" in json_data or '[' in json_data:
        # json string
        result = ast.literal_eval(json_data)
        return result
    else:
        # file path, open
        with open(json_data, 'r', encoding='UTF-8') as file:
            data = file.read()
            result = ast.literal_eval(data)
        return result
    

def create_get_request(url, params={}, headers={}):
    """
        Creates a request to the url and returns the response in a dictionary
        Args:
            url (string): The requestes url
            params (dict): The parameters in the Url
            headers (dict): The values in the ehader
        Returns:
            dict: Response data in dictionary form
    """
    try:
        r = requests.get(url=url, params=params, headers=headers)
        r.raise_for_status()
        return r.json()
    except requests.exceptions.HTTPError as err:
        print_error('get request to ' + url +' failed ' + str(err))
        return None


def create_post_request(url, body, params={}, headers={}):
    """
        Creates a request to the url and returns the response in a dictionary
        Args:
            url (string): The requestes url
            params (dict): The parameters in the Url
            headers (dict): The values in the ehader
            body ()
        Returns:
            dict: Response data in dictionary form
    """
    try:
        r = requests.post(url=url, params=params, headers=headers, data=body)
        r.raise_for_status()
        return r.json()
    except requests.exceptions.HTTPError as err:
        print_error('post request to ' + url +' failed ' + str(err))
        return None


#
#   Created by Ryan McDermott
#   Created on 3/16/2022
#

import sys
from google.cloud import bigquery
from google.api_core.exceptions import NotFound
from indxdatalaketools import Helpers
from indxdatalaketools.DataLakeGlobals import DATASET_OWNER
from indxdatalaketools.Helpers import print_error
from google.cloud.bigquery_storage_v1 import types
from google.cloud.bigquery_storage_v1 import writer
from google.cloud.bigquery.enums import EntityTypes


class Wrapper:
    '''
        Wrapper for Google Cloud Platform Big Query python SDK
    '''

    __google_big_query_client = None
    file_format_map = {
        'AVRO':bigquery.SourceFormat.AVRO,
        'CSV':bigquery.SourceFormat.CSV
        }
    
    
    def __init__(self, google_clients):
        self.__google_big_query_client = google_clients.big_query_client


    def insert_to_bq_table(self, bq_storage_client, serialized_row, request_template, stream_name, parent):
        '''
            Function that inserts data into a big query table
            Args:
                bq_client (google.cloud.bigquery.client): The client that interacts
                    with bigquery api
                serialized_row: The serialized proto buffer row
                request_template: The template the request needs to follow
                stream_name (string): The name of the stream
                parent: The parent
            Returns:
                repsonse: The response from the API call
        '''
        append_rows_stream = writer.AppendRowsStream(bq_storage_client, request_template)
        proto_rows      = types.ProtoRows()
        proto_rows.serialized_rows.append(serialized_row)

        # Set an offset to allow resuming this stream if the connection breaks.
        # Keep track of which requests the server has acknowledged and resume the
        # stream at the first non-acknowledged message. If the server has already
        # processed a message with that offset, it will return an ALREADY_EXISTS
        # error, which can be safely ignored.
        #
        # The first request must always have an offset of 0.
        request                 = types.AppendRowsRequest()
        request.offset          = 0
        proto_data              = types.AppendRowsRequest.ProtoData()
        proto_data.rows         = proto_rows
        request.proto_rows      = proto_data
        request.write_stream    = stream_name

        try:
            response_future_1 = append_rows_stream.send(request)
        except Exception as e:
            print("ERROR:" + str(e))
            return None

        append_rows_stream.close()

        bq_storage_client.finalize_write_stream(name=stream_name)

        batch_commit_write_streams_request                  = types.BatchCommitWriteStreamsRequest()
        batch_commit_write_streams_request.parent           = parent
        batch_commit_write_streams_request.write_streams    = [stream_name]
        bq_storage_client.batch_commit_write_streams(batch_commit_write_streams_request)

        return response_future_1


    def batch_load_from_gcs(self, gcs_uri, table_destination, file_format):
        '''
            Batch loads a file from gcs into big query
            Args:
                gcs_uri (string): the gcs uri of the file we wish to load into big query
                table_destination (string): The destination table
                file_format (string): the file format we are loading into big query 
                    (AVRO, CSV)
            Returns:
                boolean: True if the operation was successful, False if otherwise
        '''
        job_config = bigquery.LoadJobConfig(
            use_avro_logical_types=True,
            write_disposition=bigquery.WriteDisposition.WRITE_TRUNCATE,
            source_format=self.file_format_map[file_format])
        load_job = self.__google_big_query_client.load_table_from_uri(
            gcs_uri, table_destination, job_config=job_config
        )  # Make an API request.
        try:
            load_job.result()
            return True
        except Exception as e:
            print_error(str(e))
            return False

    

    def create_dataset_if_not_exist(self, dataset_id, location):
        '''
            Creates the client's Dataset if it does not already exist
            Args:
                dataset_id (string): The id of the dataset for the table
                location (string): The location of the dataset
            Returns:
                boolean: True if the dataset already exists or was created, False if otherwise
        '''
        dataset             = self.get_dataset(dataset_id)

        if dataset is None:
            return False

        dataset.location    = location
        # Check if dataset exists
        if self.check_if_dataset_exists(dataset_id):
            print("Dataset {} already exists, skipping this step".format(dataset_id))
            return True

        print("Creating dataset: " + dataset_id)
        self.create_dataset(dataset)

        return self.check_if_dataset_exists(dataset_id)


    def get_dataset(self, dataset_id):
        '''
            Gets the dataset in bigquery
            Args:
                dataset_id (string): The dataset Id we wish to retrieve or create
            Returns:
                google.cloud.bigquery.Dataset
        '''
        try:
            return bigquery.Dataset(dataset_id)
        except Exception as e:
            Helpers.print_error('Could not retreive dataset ' + dataset_id + " " + str(e))
            return None

        

    def create_table_if_not_exist(self, dataset_id, table_name, schema):
        '''
            Creates the client's PATIENT table if it does not already exist
            Args:
                dataset_id (string): The id of the dataset for the table
                table_name (string): The name of the table we wish to create
            Returns:
                boolean: True if the table was created or already existed, false
                    if table creation failed
        '''
        # create table object  
        table_id    = dataset_id + "." + table_name
        table       = bigquery.Table(table_id, schema=schema)

        if self.check_if_table_exists(table_id):
            print("Table {} already exists, skipping this step".format(table_id))
            return True
                    
        print("Creating table " + table_id)
        table = self.create_table(table)
            
        return self.check_if_table_exists(table_id)
        

    def create_table(self, table):
        '''
            Creates Big Query Table
            Args:
                table (bigquery.Table): The table we wish to create
            returns:
                bigquery.Table: The created Table
        '''
        try:
            return self.__google_big_query_client.create_table(table)
        except Exception as e:
            Helpers.print_error("Could not connect to GCS client " + str(e))
            sys.exit(1)


    def create_dataset(self, dataset):
        '''
            Creates Big Query Dataset
            Args:
                dataset (bigquery.Dataset): The dataset we wish to create
            returns:
                None
        '''
        try:
            self.__google_big_query_client.create_dataset(dataset, timeout=30)
        except Exception as e:
            Helpers.print_error("Could not connect to GCS client " + str(e))
            sys.exit(1)


    def check_if_table_exists(self, table_id):
        '''
            Checks if the table Id passed exists on GCP
            Args:
                dataset_id (string): The table {PROJECT_ID}.{CLIENT_UUID}.PATIENTS
            Returns:
                boolean: True if the table exists, false if other wise
        '''
        try:
            self.__google_big_query_client.get_table(table_id)  # Make an API request.
            return True
        except NotFound:
            return False
        except Exception as e:
            print(str(e))
            return False


    def check_if_dataset_exists(self, dataset_id):
        '''
            Checks if the dataset ID passed exists on GCP
            Args:
                dataset_id (string): The dataset {PROJECT_ID}.{CLIENT_UUID}
            Returns:
                boolean: True if the datset exists, false if other wise
        '''
        try:
            self.__google_big_query_client.get_dataset(dataset_id)  # Make an API request.
            return True
        except NotFound:
            return False
        except Exception as e:
            print(str(e))
            return False


    def delete_table(self, table_id):
        '''
            Deletes a table
            Args:
                table_id (string): The table id (your-project.your_dataset.your_table)
            Returns:
                boolean: True if the table was deleted false if otherwise
        '''
        try:
            self.__google_big_query_client.delete_table(table_id)
            return True
        except Exception as e:
            print_error('Failed to Delete table: ' + table_id + " " + str(e))
            return False


    def perform_query(self, query):
        '''
            Performs a big query query
            Args:
                query (string): the query we wish to perform
            Returns:
                query_job: The job of the query 
        '''
        try:
            query_job   = self.__google_big_query_client.query(query)
            rows        = query_job.result()
            return rows
        except Exception as e:
            Helpers.print_error('Could not connect to GCS Big Query ' + str(e))
            return None


    def set_permissions_on_dataset(self, dataset_id, role, account):
        '''
            Function that sets the permissions of a dataset
            Args:
                dataset_id (string): The dataset
                role (string): The role we want the account to have
                account (string): The account we are giving permission to
            Returns:
                boolean: True if the permissions were set, False if otherwise
        '''
        dataset     = self.get_dataset(dataset_id)
        entity_type = EntityTypes.USER_BY_EMAIL
        entries     = list(dataset.access_entries)
        # need to add an owner role to the dataset to set the SA permissions
        entries.append(
            bigquery.AccessEntry(
                role='roles/bigquery.dataOwner',
                entity_type=entity_type,
                entity_id=DATASET_OWNER,
            )
        )
        entries.append(
            bigquery.AccessEntry(
                role=role,
                entity_type=entity_type,
                entity_id=account,
            )
        )
        
        dataset.access_entries = entries

        return self.update_dataset(dataset, ["access_entries"])


    def update_dataset(self, dataset, fields):
        '''
            Updates a dataset with the fields passed
            Args:
                dataset (bigquery.Dataset): The dataset we wish to update
                fields (list): list of fields we wish to update
            Returns:
                boolean: True if the dataset was updated
        '''
        try:
            self.__google_big_query_client.update_dataset(dataset, fields)
            return True
        except Exception as e:
            print_error(e)
            return False




#
#   Created by Ryan McDermott
#   Created on  2/24/2022
#

import re
import sys
import cv2

import numpy as np

from indxdatalaketools import Helpers
from indxdatalaketools.ClientTools.PatientsTable import Insert
from indxdatalaketools.ClientTools.Changelog import Insert as ChnglogInsert
from indxdatalaketools.DataStandardizer import PatientTable


class Client:
    '''
        Class that contains all the functionality to upload a file
        to a data lake
    '''

    client_uuid                         = ''
    text_file_list                      = ["txt", "tsv", "csv"]

    __google_storage_client             = None
    __patients_table_client             = None
    __changelog_client                  = None
    __data_standardizer                 = None

    def __init__(self, google_clients, client_uuid):
        '''
            Init function that sets up client credentials, uuid, 
            and google api clients used to upload files to a GCS bucket
            Args:
                google_clients          (ApiClients): object containing all Google Api Clients
                client_uuid             (string): The client's uuid
            Returns:
                None
        '''
        self.client_uuid                = client_uuid
        self.__google_storage_client    = google_clients.cloud_storage_client
        self.__patients_table_client    = Insert.Client(
            google_clients, self.client_uuid)
        self.__changelog_client         = ChnglogInsert.Client(
            google_clients, self.client_uuid)
        self.__data_standardizer        = PatientTable.Client()
                

    def close(self):
        '''
            Frees and destroys all allocated resources
            Returns:
                None
        '''
        self.__google_storage_client.close()
        self.__patients_table_client.close()


    def upload_file_to_datalake(self, modality, mrn, metadata, file_path):
        '''
            Uploads a file to the data lake and sets the metadata
            for that file
            Args:
                modality    (string): The file's modality
                mrn         (string): The mrn of the patient to who the files belongs to
                metadata    (string): The metadata that is associated with the
                    file
                file_path   (string): The path to the file we wish to upload to a 
                    datalake
            Returns:
                boolean: True if the File was successfully uploaded, False if 
                    otherwise
                
        '''
        file_contents       = self.__read_file_contents(file_path)

        # convert bmp to png
        if '.bmp' in file_path.lower():
            file_contents = self.__convert_bmp_to_png(file_contents)
            file_path = file_path.replace(file_path[len(file_path) - 3:], 'png')

        stripped_mrn        = self.__strip_leading_zeros(mrn)
        hashed_patient_mrn  = Helpers.patient_hash(self.client_uuid, [stripped_mrn])[0]
        hashed_file_name    = Helpers.hash_contents_to_file_name(file_path, file_contents)
        blob_name           = modality + "/" + hashed_patient_mrn + "/" + hashed_file_name
        bucket              = self.__google_storage_client.bucket(self.client_uuid)
        blob                = bucket.blob(blob_name)

        if Helpers.try_blob_exists(blob):
            Helpers.print_error("File" + blob.name + " already exists")
            return False

        if not self.__create_blob_with_metadata(blob, file_contents, metadata):
            Helpers.print_error("Could not upload file " + file_path + " to datalake")
            return False

        if hashed_patient_mrn != 'UNKNOWN'and \
            not self.__insert_data_to_patients_table(stripped_mrn, hashed_patient_mrn):
            Helpers.print_error("Could not insert data to PATIENTS table")
            return False

        blob.metadata = metadata
        return self.__changelog_client.insert_to_changelog_table(blob, 'CREATED')


    def __convert_bmp_to_png(self, file_bytes):
        '''
            Converts the BMP bytes string into a png bytes string
            Args:
                byte (string): The string of bytes we wish to convert
            Returns:
                string: The converted bytes in string format
        '''
        nparr       = np.frombuffer(file_bytes, dtype=np.uint8)
        img         = cv2.imdecode(nparr, cv2.IMREAD_COLOR)
        png_img     = cv2.imencode('.png', img)[1]
        data_encode = np.array(png_img)
        byte_encode = data_encode.tobytes()

        return byte_encode


    def __strip_leading_zeros(self, string_to_strip):
        '''
            strips the leading '0' from a string
            Args:
                string_to_strip (string): Te string we wish to strip
            Returns:
                string: The string with leading '0' stripped
        '''
        stripped_string = re.sub("^0+", "", string_to_strip)

        return stripped_string

    
    def __insert_data_to_patients_table(self, mrn, hashed_patient_mrn):
        '''
            Inserts the file's patient to the patient table
            Args:
                mrn (string): the patients mrn
                hashed_patient_mrn (stirng): the patient id
        '''
        patient_data = {"MRN":mrn, 
                        "CLIENT_ID":self.client_uuid,
                        "PATIENT_ID":hashed_patient_mrn
                        }

        patient_data = self.__data_standardizer.standardize(patient_data)
            
        if self.__patients_table_client.insert_data_to_patients_table(patient_data):
            return True

        Helpers.print_error('Failed to upload ' + str(patient_data) + ' to patients table')
        return False


    def __create_blob_with_metadata(self, blob, file_contents, metadata):
        '''
            Creates the blob in GCS and adds metadata to it
            Args:
                blob (google.cloud.storage.blob): The blob we want to upload
                file_contents (bytes): The bytes of the file we wish to upload
                metadata (dict): The files Metadata
            Returns:
                boolean: True if the operation was successful, False if otherwise
        '''
        # add metadata to blob and save in bucket
        blob.metadata = metadata
        
        if not self.__try_upload_blob(blob, file_contents):
            return False

        if Helpers.try_blob_exists(blob):
            print('successfully uploaded file to datalake')
            return True

        return False


    def __read_file_contents(self, file_path):
        """
            Reads file from the file path, has the capabilities to read from local file,
            network file, and url. The file will be decoded to UTF-8 and remove carraige returns
            "\r" to match LF encoding.

            Args:
                file_path: The local file path, server file path or url
            Returns:
                string: string representation of data found
        """
        
        extension       = file_path.split("/")[-1].split(".")[-1].lower()

        if "gs://" in file_path:
            # GCS BUCKET
            if extension in self.text_file_list:
                content = self.__read_gcs_blob(file_path)
                return content.replace(b'\r\n', b'\n')
            else:
                file_bytes = self.__read_gcs_blob(file_path)
                return file_bytes
        
        # LOCAL FILE
        if extension in self.text_file_list:
            content = open(file_path, 'rb').read()
            return content.replace(b'\r\n', b'\n')
        else:
            file_bytes = open(file_path, 'rb').read()
            return file_bytes
    

    def __read_gcs_blob(self, gcs_blob_path):
        '''
            Reads a blob from GCS and returns the bytes 
            Args:
                gcs_blob_path (string): The path to the blob
            Returns:
                bytes: contents of the gcs blob
        '''
        file_path           = gcs_blob_path.replace("gs://", "")
        bucket_name         = file_path.split("/")[0]
        source_blob_name    = file_path.split("/", 1)[1]
        extension           = file_path.split("/")[-1].split(".")[-1].lower()
        bucket              = self.__google_storage_client.bucket(bucket_name)
        blob                = bucket.blob(source_blob_name)

        content = blob.download_as_bytes()

        return content

    def __try_upload_blob(self, blob, contents):
        '''
            Tries to upload a blob to GCS with string contents
            Args:
                blob (google.cloud.storage.Blob): The blob we wish to upload to
                contents (string): The files contents
        '''
        try:
            blob.upload_from_string(contents)
            return True
        except Exception as e:
            Helpers.print_error("Could not upload blob " + str(e))
            return False


#
#   Created By Ryan McDermott
#   Created on 2/23/2022
#

from google.api_core.exceptions import NotFound
from indxdatalaketools import Helpers

class Validator:
    '''
        class that validates the Client Id for a command. Different commands require
        different forms of validation.
    '''
    client_uuid                         = ''
    data_lake_command                   = ''
    credentials_file_path               = ''

    __client_id_validation_functions    = None
    __google_datastore_client           = None
    __google_cloud_storage_client       = None
    __google_bigquery_client            = None


    def __init__(self, google_clients):

        self.__google_datastore_client      = google_clients.datastore_client
        self.__google_bigquery_client       = google_clients.big_query_client
        self.__google_cloud_storage_client  = google_clients.cloud_storage_client

        self.__set_up_client_id_validation_functions()


    def __set_up_client_id_validation_functions(self):
        '''
            Sets up the mapping from the command passed to the function that will validate
            the client_id
            Args:
                None
            Returns:
                None
        '''
        self.__client_id_validation_functions = {
            'datalake': self.__validate_client_is_in_datastore,
            'snapshot': self.__validate_primary_and_snapshot_buckets,
            'patients-table': self.__validate_dataset_and_table,
            'file': self.__validate_primary_bucket,
            'file update': self.__validate_primary_bucket
        }
        

    def close(self):
        '''
            Frees and destroys all allocated resources
            Returns:
                None
        '''
        self.__google_cloud_storage_client.close()
        self.__google_datastore_client.close()
        self.__google_bigquery_client.close()


    def validate(self, client_uuid, data_lake_command):
        '''
            Validates the client ID
            Returns:    
                boolean: True if the client id is valid, False if otherwise
        '''
        self.client_uuid = client_uuid
        return self.__client_id_validation_functions[data_lake_command]()


    def __validate_client_is_in_datastore(self):
        '''
            Verifies the client id exists in Datastore      
            Returns:
                boolean: True if the client id is valid, False if otherwise
        '''
        key                 = self.__google_datastore_client.key('Client', self.client_uuid)
        entity              = self.__google_datastore_client.get(key=key)

        # Validate Client entity exists, exit if not
        if entity is None:
            Helpers.print_error(self.client_uuid + " does not exist in datastore: not a valid CLIENT_ID")
            return False
        
        return True


    def __validate_primary_and_snapshot_buckets(self):
        '''
            Verifies the client id primary bucket exists and well as the snapshot bucket
            Returns:
                boolean: True if the client id is valid, False if otherwise
        '''
        
        data_lake_snapshot_bucket   = self.__google_cloud_storage_client.\
            bucket(self.client_uuid + '-snapshots')

        if not Helpers.try_bucket_exists(data_lake_snapshot_bucket):
            Helpers.print_error('DataLake snapshot bucket does not exist: Bad Request')
            return False

        return self.__validate_primary_bucket()


    def __validate_primary_bucket(self):
        '''
            Verifies the client id primary bucket exists and well as the snapshot bucket
            Returns:
                boolean: True if the client id is valid, False if otherwise
        '''
        data_lake_bucket            = self.__google_cloud_storage_client.\
            bucket(self.client_uuid)

        if not Helpers.try_bucket_exists(data_lake_bucket):
            Helpers.print_error('DataLake bucket does not exist: Bad Request')
            return False

        return True

    
    def __validate_dataset_and_table(self):
        '''
            Verifies the client id dataset and PAETIENTS table exists
            Returns:
                boolean: True if the client id is valid, False if otherwise
        '''
        client_id_conformed_name    = self.client_uuid.replace('-','_')
        dataset_id                  = 'indx-data-services.'+ client_id_conformed_name
        table_id                    = dataset_id + '.PATIENTS'

        if not self.__check_if_dataset_exists(dataset_id):
            Helpers.print_error('Dataset for clientId does not exist: Bad Request')
            return False

        if not self.__check_if_table_exists(table_id):
            Helpers.print_error('PATIENTS table does not exist: Bad Request')
            return False

        return True


    def __check_if_table_exists(self, table_id):
        '''
            Checks if the table Id passed exists on GCP
            Args:
                dataset_id (string): The table {PROJECT_ID}.{DATASET}.PATIENTS
            Returns:
                boolean: True if the table exists, false if other wise
        '''
        try:
            self.__google_bigquery_client.get_table(table_id)  # Make an API request.
            return True
        except NotFound:
            return False

    def __check_if_dataset_exists(self, dataset_id):
        '''
            Checks if the dataset ID passed exists on GCP
            Args:
                dataset_id (string): The dataset {PROJECT_ID}.{CLIENT_UUID}
            Returns:
                boolean: True if the datset exists, false if other wise
        '''
        try:
            self.__google_bigquery_client.get_dataset(dataset_id)  # Make an API request.
            return True
        except NotFound:
            return False


#
#   Created by Ryan McDermott
#   Created on 5/12/2022
#

from indxdatalaketools import GoogleClients
from indxdatalaketools.PropertiesDBApi import Properties
from indxdatalaketools.ArgumentValidation import Argument

from indxdatalaketools.DataStandardizer import FileMetadata, PatientTable
from indxdatalaketools.ClientTools.Files import Upload
from indxdatalaketools.ClientTools.PatientsTable import Insert
from indxdatalaketools.DocumentClassifierApi import DocumentClassifierApi
from indxdatalaketools.GoogleClientWrappers import CloudStorage
from indxdatalaketools import Helpers
import warnings

class Client:

    client_uuid                         = ''

    __argument_validator                = None
    __google_clients                    = None
    __upload_file                       = None

    def __init__(self, client_uuid, credentials_file_path=None):
        '''
            Init function that sets up client credentials, uuid, 
            and File operations API
            Args:
                credentials_file_path   (string): The path to a Service Account key json file
                client_uuid             (string): The client's uuid
            Returns:
                None
        '''
        self.client_uuid                = client_uuid

        self.__google_clients           = GoogleClients.ApiClients(credentials_file_path)
        self.__argument_validator       = Argument.Validator(self.__google_clients)
        self.__upload_file              = Upload.Client(self.__google_clients, client_uuid)
        self.__insert_patient           = Insert.Client(self.__google_clients, client_uuid)
        self.__document_classifier      = DocumentClassifierApi.DocClassifierClient(self.__google_clients)
        self.__cloud_storage_wrapper    = CloudStorage.Wrapper(self.__google_clients)


    def upload_file(self, modality, mrn, metadata, file_path):
        '''
            Top level function that uploads a file to a Datalake, This will aslo add the 
            Patient data to the PATIENTS table if applicable
            Args:
                modality    (string): The modality of the file
                mrn         (string): The mrn of who the file belongs to
                metadata    (dict|string|file_path): Any metadata belonging to the file
                file        (string): The file path of the file we wish to upload
            Returns:
                boolean: True if the operation was a success, False if otherwise
        '''
        # validate arguments
        arguments           = {
            'client_id':self.client_uuid, 
            'modality':modality,
            'mrn':mrn,
            'file_metadata':metadata,
            'file_path':file_path}
        data_lake_command   = 'file'
        
        Properties.Properties.instance(self.__google_clients)

        if not self.__argument_validator.validate_all_arguments(arguments, data_lake_command):
            return False

        # standardize the metadata
        file_metadata_standardizer = FileMetadata.Standardizer(arguments)
        file_metadata = file_metadata_standardizer.standardize(metadata)

        # run the document classifier, check if modality matches predicted modality
        classified_modality = self.__document_classifier_operations(mrn, file_path)
        if not classified_modality:
            warnings.warn("Request to document classifier failed.")
        elif classified_modality != modality:
            Helpers.print_error("Modality may be incorrect. Modality might be " + classified_modality)

        # run command
        return self.__upload_file.upload_file_to_datalake(
            modality, mrn, file_metadata, file_path)


    def __document_classifier_operations(self, mrn, file_path):
        '''
            function that handles operations required to run the document classifier. This includes uploading file from local to 
            a staging folder on the clients data lake, running the document classifier with this uploaded blob and then deleting the blob
            args:
                mrn: str - input mrn
                file_path: str - local file path of input file
            return:
                classified_modality: str - modaity prediction of the document classifier
        '''
        blob = self.__cloud_storage_wrapper.upload_local_file_to_gcs_bucket(mrn, file_path, "staging_folder", self.client_uuid)
        if not blob:
            return blob

        gcs_file_path = "gs://"+blob.bucket.name+"/"+blob.name
        classified_modality, classified_file_type = self.__send_request_to_document_classifier(gcs_file_path)
        
        result = Helpers.delete_blob(blob)
        if not result:
            warnings.warn("Error faced when trying to delete blob from staging_folder")
        return classified_modality
        

    def __send_request_to_document_classifier(self, gcs_file_path):
        '''
            runs the document classifier
            args:
                gcs_file_path: str gcs path of file in staging folder
            return:
                classified_modality: str - modality prediction of the document classifier
                classified_file_type: str - file type prediction of the document classifier
        '''
        try:
            return self.__document_classifier.get_modality_and_file_type(gcs_file_path)
        except Exception as e:
            print("Error when trying to run document classifier. " + str(e))
            return False, False
        

    def upload_patient(self, patient_data):
        '''
            Top level function that uploads a patient to a PATIENT table, This will also update missing
            patient data but not overwrite it.
            Args:
                patient_data (dict): Dictionary containing the data to insert a patient to the patient table
            Returns:
                boolean: True if the operation was a success, False if otherwise
        '''
        # verify inputs
        arguments           = {'client_id':self.client_uuid, 'patient_data':patient_data}
        data_lake_command   = 'patients-table'

        Properties.Properties.instance(self.__google_clients)

        if not self.__argument_validator.validate_all_arguments(arguments, data_lake_command):
            return False

        # clean data
        data_standardizer   = PatientTable.Client()
        patient_data        = data_standardizer.standardize(patient_data)

        # run the command
        return self.__insert_patient.insert_data_to_patients_table(patient_data)


#
#   Created by Ryan McDermott
#   Created on 2/24/2022
#

import datetime
import re
from indxdatalaketools import Helpers
from indxdatalaketools.ArgumentValidation import ValidationHelpers
from indxdatalaketools.DataLakeGlobals import TIME_REGEX
from indxdatalaketools.PropertiesDBApi import Properties

IN_METADATA = ' in metadata'

class Validator:
    '''
        Class that validates the File metadata fields passed by the client
    '''
    client_uuid                         = ''
    data_lake_command                   = ''
    arguments                           = {}
    metadata                            = {}
    
    __metadata_validation_functions     = None
    __google_storage_client             = None

    def __init__(self, google_clients):
        
        
        self.__google_storage_client    = google_clients.cloud_storage_client

        self.__set_up_metadata_validation_functions()


    def __set_up_metadata_validation_functions(self):
        '''
            Sets up the mapping from the arguments passed in patient_data 
            to the function that will validate the given field
            Args:
                None
            Returns:
                None
        '''
        self.__metadata_validation_functions = {
            'CLIENT_ID': self.__validate_client,
            'PATIENT_ID': self.__validate_patient_id,
            'MODALITY': self.__validate_modality,
            'FILE_PATH': self.__validate_file_path,
            'DATE_OF_SERVICE': ValidationHelpers.validate_date,
            'TIME_OF_SERVICE': self.__validate_time_of_service,
            'LOCATION_OF_SERVICE': self.__validate_location_of_service,
            'SOURCE': self.__validate_source,
            'ORIGINATOR': self.__validate_originator
        }

    
    def close(self):
        '''
            Frees and destroys all allocated resources
            Returns:
                None
        '''
        self.__google_storage_client.close()


    def validate(self, arguments, data_lake_command):
        '''
            Validates the metadata
            Args:
                metadata: structure or file containing one or more patients
                    data
            Returns:
                boolean: True if all Patients have valid data, False if otherwise
        '''
        self.client_uuid        = arguments['client_id']
        self.arguments          = arguments
        self.data_lake_command  = data_lake_command
        json_structure          = Helpers.determine_json_structure(arguments['file_metadata'])
        self.metadata           = json_structure

        return self.__validate_metadata_json(json_structure)

    
    def __validate_metadata_json(self, metadata):
        '''
            Verifies that the json passed contains valid attributes
            Args:
                None
            Returns:
                boolean: True if the patient data is valid, False if otherwise
        '''
        for field in metadata:
            if field not in self.__metadata_validation_functions:
                continue
            if not self.__validate_field_in_metadata(field, metadata):
                return False

        return True


    def __validate_field_in_metadata(self, field, metadata):
        '''
            Validates a given fields in the patient data json structure using the mapping
            stored in the .__metadata_validation_functions dictionary
            Args:
                field           (string): The field we are going to validate
                patient_data    (dict): The dictionary of the patient data
            Returns:
                boolean: True if the field value is valid, False if otherwise
        '''
        return self.__metadata_validation_functions[field](metadata[field])
    

    def __validate_client(self, client):
        '''
            Validates the Client ID in the metadata
            Args:
                client (string): The client ID
            Returns:
                boolean: True, of the client is valid, False if otherwise
        '''
        if self.data_lake_command == 'file' and client == self.client_uuid:
            return True

        if self.data_lake_command == 'file update' and self.__check_if_bucket_exists(client):
            return True

        Helpers.print_error('Invalid CLIENT_ID' + client + IN_METADATA)
        return False


    def __check_if_bucket_exists(self, client):
        '''
            Checks if the client bucket exists, This only needs to be checked if we are
            switching patients' files between clients due to a messup
            Args:
                client (string): The client ID
            Returns:
                boolean: True, of the client is valid, False if otherwise
        '''
        bucket = self.__google_storage_client.bucket(client)
        result = Helpers.try_bucket_exists(bucket)

        if result:
            return True

        Helpers.print_error("datalake " + client + " does not exist: Bad Request")
        return False


    def __validate_patient_id(self, patient_id):
        '''
            Validates the Patient Id:
            Args:
                patient_id (string): The patient id
            Returns:
                boolean: True if the patient id is valid, False if otherwise
        '''
        expected_patient_id = Helpers.patient_hash(self.arguments['client_id'],
            [self.arguments['mrn']])[0]
        
        if patient_id != expected_patient_id:
            Helpers.print_error('Invalid PATIENT_ID ' + patient_id + IN_METADATA)
            return False

        return True


    def __validate_modality(self, modality):
        '''
            Validates the modality:
            Args:
                modality (string): The modality
            Returns:
                boolean: True if the modality is valid, False if otherwise
        '''
        if modality in Properties.Properties.instance().modalities:
            return True

        Helpers.print_error('Invalid MODALITY ' + modality + IN_METADATA)
        return False


    def __validate_file_path(self, gcs_file_path):
        '''
            Validates the gcs file path
            Args:
                gcs_file_path (string): the path in GCS where the file exists
            Returns:
                boolean: True if the file path matches with patient_id and
                    modality
        '''
        expected_patient_id = Helpers.patient_hash(self.arguments['client_id'], 
            [self.arguments['mrn']])[0]
        expected_file_path  = self.arguments['modality'] + '/' + expected_patient_id

        if gcs_file_path == expected_file_path:
            return True

        Helpers.print_error('file path ' + gcs_file_path + ' does not match the expected file path' + 
            expected_file_path + ' created from the arguments')
        return False


    # TO-DO update location of service validation after we know what to expect
    def __validate_location_of_service(self, location_of_service):
        '''
            validates the location of service
             Args:
                location_of_service (string): The location of service
            Returns:
                boolean: True if time of service is valide, false if otherwise
        '''
        return True


    # TO-DO update source validation after we know what to expect
    def __validate_source(self, source):
        '''
            validates the source
            Args:
                source (string): The source
            Returns:
                booleean: True if the source is valid, False if otherwise
        '''
        return True


    # TO-DO update originator validation after we know what to expect
    def __validate_originator(self, originator):
        '''
            validates the originator
            Args:
                source (string): The originator
            Returns:
                booleean: True if the originator is valid, False if otherwise
        '''
        return True


    def __validate_time_of_service(self, time_of_service):
        '''
            Validates the time of service
            Args:
                time_of_service (string): The data of service
            Returns:
                boolean: True if time of service is valide, false if otherwise
        '''
        if re.fullmatch(TIME_REGEX, time_of_service) is None:
            print(time_of_service + ' does not match time regex: not a valid time')
            return False

        #check valid time
        hour    = int(time_of_service[1:3])
        minute  = int(time_of_service[3:5])
        second  = int(time_of_service[5:7])

        if self.__is_valid_time(hour, minute, second):
            return True

        Helpers.print_error(time_of_service + ' is not a valid date')
        return False

    
    def __is_valid_time(self, hour, minute, second):
        '''
            checks if the year month and day creates a valid date
            Args:
                year    (string): The year
                month   (string): The month
                day     (string): the day
            Returns:
                boolean: True if a valid date is formed False if other wise
        '''
        try:
            datetime.time(hour, minute, second)
            return True
        except:
            return False

    


#
#   Created by Ryan McDermott
#   Created on 2/22/2022
#

import datetime
import sys
import hashlib
import json
from indxdatalaketools import Helpers

HL7_SEX_MAP         = {"F":"female","M":"male","O":"other","U":"unknown",
                        "female":"female","male":"male","other":"other","unknown":"unknown"}


class Client():
    '''
        Creates Big query readable dictionary for the PATIENT table
    '''
    def __init__(self):
        pass
    

    def standardize(self, patient_data):
        '''
            Standardizes the patient json data to a form used by the 
            patients table API
            Args:
                patient_data (string|dict): Converts the data to a dictionary used by 
                PATIENT table API
            Returns:
                dict: Dictionary containing all fields in PATIENT schema, if the fields does
                    does not exist it is set to NONE
        '''
        json_structure = Helpers.determine_json_structure(patient_data)

        if isinstance(json_structure, list):
            return self.__standardize_multiple_patients_data_json(json_structure)

        return self.__standardize_patient_data(json_structure)


    def __standardize_multiple_patients_data_json(self, json_structure):
        '''
            Verifies multiple patients data passed by the client
            Args:
                json_structure (list[dict]): list of patient data
            returns:
                boolean: True if all patients' data is valid
        '''
        standardized_patients = []

        for patient in json_structure:

            standardized_patient_data = self.__standardize_patient_data(dict(patient))
            standardized_patients.append(standardized_patient_data)

        return standardized_patients

    
    def __standardize_patient_data(self, patient_data):
        '''
            converts all missing fields in patient data to None, updates the date format,
            and removes all extra fields in patient data
            Args:
                patient_data (dict):Dictionary of patient Data
            Returns:
                patient_data (dict):Dictionary of patient Data
        '''
        patient_data = self.__set_patient_id(patient_data)

        patient_data = self.__update_missing_fields_to_none(patient_data)

        patient_data = self.__remove_all_extra_fields_from_json(patient_data)

        patient_data = self.__convert_hl7_sex_code(patient_data)

        patient_data = self.__transform_date(patient_data)

        return self.__create_checksum(patient_data)
        

    def __set_patient_id(self, patient_data):
        '''
            checks if the PATIENT_ID exists if not it is created, if MRN and CLIENT_ID
            do not exist the program exits
            Args:
                patient_data (dict): Dictionary of patient Data
            Returns:
                patient_data (dict): Dictionary of patient Data
        '''
        if "PATIENT_ID" not in patient_data:
            patient_data["PATIENT_ID"] = Helpers.patient_hash(patient_data["CLIENT_ID"], 
            [patient_data["MRN"]])[0]

        return patient_data


    def __update_missing_fields_to_none(self, patient_data):
        '''
            converts all missing fields in patient data to None, updates the date format,
            and removes all extra fields in patient data
            Args:
                patient_data (dict):Dictionary of patient Data
            Returns:
                patient_data (dict):Dictionary of patient Data
        '''
        patients_schema     = ["CLIENT_ID","MRN","LASTNAME",
        "FIRSTNAME","PATIENT_ID","MIDDLENAME","DOB","SEX","RACE" ]

        for key in patients_schema:
            if key not in patient_data:
                patient_data[key] = None

        return patient_data

    def __convert_hl7_sex_code(self, patient_data):
        '''
            Converts the HL7 sex code if it is not in the right format 
            Args:
                patient_data (dict):Dictionary of patient Data
            Returns:
                patient_data (dict):Dictionary of patient Data
        '''
        if patient_data['SEX'] is None:
            return patient_data

        patient_data['SEX'] = HL7_SEX_MAP[patient_data['SEX']]
        return patient_data


    # Reviewers: Is there a better way to implement this logic?
    # cannot loop through a list and remove items while you loop
    def __remove_all_extra_fields_from_json(self, patient_data):
        '''
            Function that removes all fields from the json dictionary if
            they are not part of the PATIENTS table schema
            Args:
                patient_data (dict):Dictionary of patient Data
            Returns:
                None
        '''
        remove_arr          = []
        patients_schema     = ["CLIENT_ID","MRN","LASTNAME",
        "FIRSTNAME","PATIENT_ID","MIDDLENAME","DOB","SEX","RACE" ]

        for k in patient_data:
            if k not in patients_schema:
                remove_arr.append(k)
               
        for k in remove_arr:
            patient_data.pop(k, None)

        return patient_data


    def __create_checksum(self, patient_data):
        '''
            Create an MD5 checksum field in the dictionary 
            Args:
                patient_data (dict): The dictionary containing the patients data
            Returns:
                dict: The updated dictionary
        '''
        dhash = hashlib.md5()
        encoded = json.dumps(patient_data, sort_keys=True).encode()
        dhash.update(encoded)
        patient_data['CHECKSUM'] = dhash.hexdigest()
        return patient_data

    
    def __transform_date(self, patient_data):
        '''
            Create a data time using the passed paramaters
            Args:
                patient_data (dict): The dictionary containing the patients data
            Returns:
                boolean: True if the datetime was successfully created, False
                    if otherwise
        '''
        if patient_data['DOB'] is None:
            return patient_data

        year    = int(patient_data["DOB"][:4])
        month   = int(patient_data["DOB"][4:6])
        day     = int(patient_data["DOB"][6:8])

        patient_data['DOB'] = self.__create_datetime_date_from_dob(year, month, day)

        return patient_data


    def __create_datetime_date_from_dob(self, year, month, day):
        '''
            Creates a date time data object from the year month and day passed
            Args:
                year    (string): The year
                month   (string): The month
                day     (string): The day
            Returns:
                The correctly Formatted date
        '''
        try:
            date = datetime.date(year, month, day)
            return str(date)        
        except:
            Helpers.print_error("Date is not a valid date: Bad Request")
            sys.exit(1)


#
#   Created By Ryan McDermott
#   Created On 2/17/2022
#
#   Top level file that hold all the logic for calling the executables. This
#   includes argument validations and implementing the sub modules for each 
#   command

import sys
import click
from google.cloud import storage
from indxdatalaketools import __version__
from indxdatalaketools.DataLakeOperations import DataLakeOps
from indxdatalaketools.DataLakeOperations import ArchiveServiceOps
from indxdatalaketools.DataLakeOperations import SnapshotOps
from indxdatalaketools.DataLakeFileOperations import PatientsTableOps
from indxdatalaketools.DataLakeFileOperations import FileOps
from indxdatalaketools.Helpers import print_error

@click.group()
@click.version_option(__version__)
def main():
    """ ALL DATA LAKE OPERATIONS """

#####
#   Data Lake Commands
#####
@main.group()
def datalake():
    """Commands for datalake operations"""


@datalake.command('create')
@click.option('--service_account', default=None, help='The path to the service account key json file')
@click.argument('client_id')
def create(client_id, service_account):
    """\b\nCreates a datalake for a given Client ID and all supporting infrastrcuture

    Arguments: 

        CLIENT_ID:   The specific UUID4 for a client.
    """
    data_lake_client    = DataLakeOps.Client(client_id, service_account)
    result              = data_lake_client.create_data_lake()
    sys.exit(int(not result))

@datalake.command('restore')
@click.option('--service_account', default=None, help='The path to the service account key json file')
@click.argument('client_id')
@click.argument('restore_date')
@click.argument('archive_bucket')
def restore_datalake(client_id, restore_date, archive_bucket, service_account):
    """\b\nRestores the datalake for a given Client ID up to a specified restore date

    Arguments: 

        CLIENT_ID:   The specific UUID4 for a client.
        RESTORE_DATE: The date in YYYYMMDD format
        ARCHIVE_BUCKET: The bucket you wish to restore from
    """
    
    archive_client    = ArchiveServiceOps.Client(client_id, service_account)
    data_lake_client  = DataLakeOps.Client(client_id, service_account)

    if not data_lake_client.create_data_lake():
        print_error('Failed to recreate datalake ' + client_id)
        sys.exit(1)
    
    result            = archive_client.restore_data_lake(restore_date, archive_bucket)
    sys.exit(int(not result))



#####
#   Snapshot Commands
#####
@main.group()
def snapshot():
    """Commands for datalake snapshot operations"""


@snapshot.command('datalake')
@click.option('--service_account', default=None, help='The path to the service account key json file')
@click.argument('client_id')
def datalake_snapshot(client_id, service_account):
    """\b\nCreates a snapshot of the current specified datalake
    
    Arguments: 

        CLIENT_ID:   The specific UUID4 for a client.
    """
    # verify inputs
    snapshot_client     = SnapshotOps.Client(client_id, service_account)
    result              = snapshot_client.create_snapshot()
    sys.exit(int(not result))


@snapshot.command('patient')
@click.option('--service_account', default=None, help='The path to the service account key json file')
@click.argument('client_id')
@click.argument('patient_id')
def datalake_patient_snapshot(client_id, patient_id, service_account):
    """\b\nCreates a snapshot of the current specified datalake
    
    Arguments: 
        CLIENT_ID:      The specific UUID4 for a client we wish to take a snapshot for
        PATIENT_ID:     The ID of the patient we wish to take a snapshot of
    """
    # verify inputs
    snapshot_client     = SnapshotOps.Client(client_id, service_account)
    result              = snapshot_client.create_patient_snapshot(patient_id)
    sys.exit(int(not result))


#####
#   PATIENTS Table Commands
#####
@main.group()
def patients_table():
    """Commands for PATIENTS table operations"""


@patients_table.command('upload')
@click.option('--service_account', default=None, help='The path to the service account key json file')
@click.argument('client_id')
@click.argument('patient_data')
def upload_patient(client_id, patient_data, service_account):
    """\b\nuploads a single, or multiple patients to the PATIENTS table
    
    Arguments: 
        
        CLIENT_ID:      The specific UUID4 for a client.
        PATIENT_DATA:   The json string, or json file representing a single, or 
                        multiple patients. The json structure is
            PATIENT_ID:             String, UUID of patient 
            CLIENT_ID (REQUIRED):   String, UUID of client / health care agency 
            MRN (REQUIRED):         String, patient’s medical record number per the client’s EMR system 
            LASTNAME:               String, patient’s last name 
            FIRSTNAME:              String, patient’s first name 
            MIDDLENAME:             String, patient’s middle name 
            DOB:                    Date, patient’s date of birth 
            SEX:                    String, patient’s administrative gender as per the HL7 code 
            RACE:                   String, patient’s race as per the HL7 code 
    """
    patients_table_client   = PatientsTableOps.Client(client_id, service_account)
    result                  = patients_table_client.insert_patients(patient_data)
    sys.exit(int(not result))


@patients_table.command('update')
@click.option('--service_account', default=None, help='The path to the service account key json file')
@click.argument('client_id')
@click.argument('patient_data')
def update_patient(client_id, patient_data, service_account):
    """\b\nuploads a single, or multiple patients to the PATIENTS table
    
    Arguments: 
        
        CLIENT_ID:      The specific UUID4 for a client.
        PATIENT_DATA:   The json string, or json file representing a single, or 
                        multiple patients. The json structure is
            PATIENT_ID:             String, UUID of patient 
            CLIENT_ID (REQUIRED):   String, UUID of client / health care agency 
            MRN (REQUIRED):         String, patient’s medical record number per the client’s EMR system 
            LASTNAME:               String, patient’s last name 
            FIRSTNAME:              String, patient’s first name 
            MIDDLENAME:             String, patient’s middle name 
            DOB:                    Date, patient’s date of birth 
            SEX:                    String, patient’s administrative gender as per the HL7 code 
            RACE:                   String, patient’s race as per the HL7 code 
    """
    patients_table_client   = PatientsTableOps.Client(client_id, service_account)
    result                  = patients_table_client.update_patients(patient_data)
    sys.exit(int(not result))


@patients_table.command('restore')
@click.option('--service_account', default=None, help='The path to the service account key json file')
@click.argument('client_id')
@click.argument('restore_date')
def restore_table(client_id, restore_date, service_account):
    """\b\nuploads a single, or multiple patients to the PATIENTS table
    
    Arguments: 
        
        CLIENT_ID:      The specific UUID4 for a client.
        RESTORE_DATE:   The date you wish to restore the PATIENTS table to
    """
    patients_table_client   = PatientsTableOps.Client(client_id, service_account)
    result                  = patients_table_client.restore_table(restore_date)
    sys.exit(int(not result))


#####
#   File Commands
#####
@main.group()
def files():
    """Commands for file operations"""

@files.command('upload')
@click.option('--service_account', default=None, help='The path to the service account key json file')
@click.argument('client_id')
@click.argument('modality')
@click.argument('mrn')
@click.argument('metadata')
@click.argument('file_path')
def upload_file(client_id, modality, mrn, metadata, file_path, service_account):
    """\b\nupload a file
    
    Arguments: 
        
        CLIENT_ID:      The specific UUID4 for a client.
        MODALITY:       The files MODALITY. acceptable parameters are:
            HVF
            OCT_RNFL
            OCT_RETINA
            AVS
        MRN:            The Patients medical record number
        METADATA:       The json string or json file of the file you wish to upload.
                        The structure is:
            CLIENT_ID (REQUIRED):   String, UUID of client / health care agency
            PATIENT_ID:             String, UUID of patient
            MODALITY:               String, The modality of the file
            FILE_PATH:              String, The path in the datalake where the file is stored
            DATE_OF_SERVICE:        String, The date of service in
            TIME_OF_SERVICE:        String, The time of service in
            LOCATION_OF_SERVICE:    String, Where the file was created
            SOURCE:                 String, The Source of the file
            ORIGINATOR:             String, The user that uploaded the file to the data lake
        FILE_PATH:      The path to the file you wish to upload
        
    """
    file_client     = FileOps.Client(client_id, credentials_file_path=service_account)
    result          = file_client.upload_file(modality, mrn, metadata, file_path)
    sys.exit(int(not result))


@files.command('delete')
@click.option('--service_account', default=None, help='The path to the service account key json file')
@click.argument('client_id')
@click.argument('modality')
@click.argument('mrn')
@click.argument('file_name')
def delete_file(client_id, modality, mrn, file_name, service_account):
    """\b\ndelete a file
    
    Arguments: 
        
        CLIENT_ID:      The specific UUID4 for a client.
        MODALITY:       The files MODALITY. acceptable parameters are:
            HVF
            OCT_RNFL
            OCT_RETINA
            AVS
        MRN:            The Patients medical record number
        FILE_NAME:      The name of the file you wish to delete
        
    """
    file_client     = FileOps.Client(client_id, service_account)
    result          = file_client.delete_file(modality, mrn, file_name)
    sys.exit(int(not result))

@files.command('update')
@click.option('--service_account', default=None, help='The path to the service account key json file')
@click.argument('client_id')
@click.argument('gsutil_uri')
@click.argument('metadata')
def update_file(client_id, gsutil_uri, metadata, service_account):
    """\b\nupdate a file's metadata
    
    Arguments: 
        
        CLIENT_ID:      The specific UUID4 for a client.
        GSUTIL_URI:     The gsutil uri of the blob you wish to update
        METADATA:       The json string or json file of the file you wish to upload.
                        The structure is:
            CLIENT_ID (REQUIRED):   String, UUID of client / health care agency
            PATIENT_ID:             String, UUID of patient
            MODALITY:               String, The modality of the file
            FILE_PATH:              String, The path in the datalake where the file is stored
            DATE_OF_SERVICE:        String, The date of service in
            TIME_OF_SERVICE:        String, The time of service in
            LOCATION_OF_SERVICE:    String, Where the file was created
            SOURCE:                 String, The Source of the file
            ORIGINATOR:             String, The user that uploaded the file to the data lake
        
    """
    file_client     = FileOps.Client(client_id, service_account)
    result          = file_client.update_file_metadata(gsutil_uri, metadata)
    sys.exit(int(not result))



#####
#   File Commands
#####
@main.group()
def patients():
    """Commands for patient operations"""

@patients.command('delete')
@click.option('--service_account', default=None, help='The path to the service account key json file')
@click.argument('client_id')
@click.argument('mrn')
def delete_patient(client_id, mrn, service_account):
    """\b\ndelete a file
    
    Arguments: 
        
        CLIENT_ID:      The specific UUID4 for a client.
        MRN:            The Patients medical record number
        
    """
    file_client     = FileOps.Client(client_id, service_account)
    result          = file_client.delete_patient(mrn)
    sys.exit(int(not result))


@patients.command('restore')
@click.option('--service_account', default=None, help='The path to the service account key json file')
@click.argument('client_id')
@click.argument('restore_date')
@click.argument('archive_bucket')
@click.argument('mrn')
def restore_patient(client_id, restore_date, archive_bucket, mrn, service_account):
    """\b\nrestore a patient to the specified date in time
    
    Arguments: 
        
        CLIENT_ID:      The specific UUID4 for a client.
        RESTORE_DATE: The date in YYYYMMDD format
        ARCHIVE_BUCKET: The specified archive bucket used for restoration
        MRN:            The Patients medical record number
    
    """
    
    archive_client    = ArchiveServiceOps.Client(client_id, service_account)
    result            = archive_client.restore_patient(restore_date, archive_bucket, mrn)
    sys.exit(int(not result))


@patients.command('delete-from-archive')
@click.option('--service_account', default=None, help='The path to the service account key json file')
@click.argument('client_id')
@click.argument('archive_bucket')
@click.argument('mrn')
def delete_patient_from_archive(client_id, archive_bucket, mrn, service_account):
    """\b\ndelete a patient from the specified archive bucket

    Arguments: 
        
        CLIENT_ID:      The specific UUID4 for a client.
        ARCHIVE_BUCKET: The specified archive bucket used for restoration
        MRN:            The Patients medical record number
    
    """
    
    archive_client    = ArchiveServiceOps.Client(client_id, service_account)
    result            = archive_client.delete_patient_in_archive(archive_bucket, mrn)
    sys.exit(int(not result))

if __name__ == '__main__':
    main()


#
#   Created by Arjun Acharya
#   Created on 7/27/2022
#

from indxdatalaketools.GoogleClientWrappers import SecretManager
from indxdatalaketools import Helpers
import json

DOCUMENT_CLASSIFIER_URL = 'https://document-classifier-gateway-27i5vb12.uc.gateway.dev/'

class DocClassifierClient:
    '''
    Class that contains all the functionality make a call to Document Classifier
    '''
    __project_id                = ''
    __secret_id                 = ''
    __api_key                   = ''
    __secret_manager_wrapper    = None

    def __init__(self, google_clients):
        self.__project_id               = 'indx-data-services'
        self.__secret_id                = 'indx-document-classifier-api-key'
        self.__secret_manager_wrapper   = SecretManager.Wrapper(google_clients)
        self.__api_key                  = self.__secret_manager_wrapper.get_secret(self.__project_id, self.__secret_id)

    def get_modality_and_file_type(self, gcs_file_path):
        '''
            creates the request to the document classifier cloud run service
            args:
                gcs_file_path: str - gcs path of file on the staging bucket
            return:
                classified_modality: str - modality prediction made by the document classifier
                classified_file_type: str - file type prediction made by the document classifier
        '''
        headers = {"Content-Type": "application/json", "x-api-key":self.__api_key}
        payload = {"gcs_file_path" : gcs_file_path}
        data = json.dumps(payload)
        data = data.encode('utf-8')
        response = Helpers.create_post_request(DOCUMENT_CLASSIFIER_URL+"predict", headers=headers, body=data)
        classified_modality = response['modality']
        classified_file_type = response['file_type']

        return classified_modality, classified_file_type


#
#   Created by Ryan McDermott
#   Created on 7/7/2022
#

from indxdatalaketools.GoogleClientWrappers import SecretManager
from indxdatalaketools import Helpers

PROPERTIES_DB_URL = 'https://properties-db-api-gate-27i5vb12.uc.gateway.dev/'

class Client:
    '''
        Class that contains all the functionality make a call to Properties
        Database URL
    '''
    __project_id                = ''
    __secret_id                 = ''
    __api_key                   = ''
    __secret_manager_wrapper    = None
    __properties_lists          = {}

    def __init__(self, google_clients):
        self.__project_id               = 'indx-data-services'
        self.__secret_id                = 'datalake-properties-db-api-key'
        self.__secret_manager_wrapper   = SecretManager.Wrapper(google_clients)
        self.__api_key                  = self.__secret_manager_wrapper.get_secret(self.__project_id, self.__secret_id)


    def populate_modality_lists(self):
        '''
            populates the poperties lists dictionary
        '''
        lists           = {}
        headers         = {"Content-Type": "application/json", "x-api-key":self.__api_key}
        response        = Helpers.create_get_request(PROPERTIES_DB_URL+"property-lists", headers=headers)
        property_lists   = response['data']
        for property_list in property_lists:
            attributes  = property_list['attributes']
            
            # do not need to include non enabled property lists
            if 'enabled' in attributes and attributes['enabled']:
                lists[attributes["list-name"]] = property_list["id"]
        
        return lists



    def get_modalities(self):
        '''
            Function that returns a list of modlaites pulled from the properties database api
            Returns:
                - list: The list of modalities 
        '''
        if self.__properties_lists == {}:
            self.__properties_lists = self.populate_modality_lists()
        
        modalities      = []
        headers         = {"Content-Type": "application/json", "x-api-key":self.__api_key}
        modalties_id    = self.__properties_lists['modalities']
        response        = Helpers.create_get_request(PROPERTIES_DB_URL+"property-lists/"+modalties_id, headers=headers)
        modality_list   = response['included']
        
        for property_value in modality_list:
            modality = property_value['attributes']["value"]
            modalities.append(modality)
        
        return modalities


    def get_race_codes(self):
        '''
            Function that returns the list of race codes pulled from the properties database api
            Returns:
                - list: The list of race codes
        '''
        if self.__properties_lists == {}:
            self.__properties_lists = self.populate_modality_lists()

        race_codes      = []
        headers         = {"Content-Type": "application/json", "x-api-key":self.__api_key}
        race_codes_id    = self.__properties_lists['race-codes']
        response        = Helpers.create_get_request(PROPERTIES_DB_URL+"property-lists/"+race_codes_id, headers=headers)
        modality_list   = response['included']
        
        for property_value in modality_list:
            race_code = property_value['attributes']["value"]
            race_codes.append(race_code)
        
        return race_codes
        




# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: table_schema.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12table_schema.proto\"\xcd\x01\n\x10table_schema_pb2\x12\x11\n\tCLIENT_ID\x18\x01 \x02(\t\x12\x12\n\nPATIENT_ID\x18\x02 \x02(\t\x12\x0b\n\x03MRN\x18\x03 \x02(\t\x12\x11\n\tFIRSTNAME\x18\x04 \x01(\t\x12\x10\n\x08LASTNAME\x18\x05 \x01(\t\x12\x12\n\nMIDDLENAME\x18\x06 \x01(\t\x12\x0b\n\x03SEX\x18\x07 \x01(\t\x12\x0c\n\x04RACE\x18\x08 \x01(\t\x12\x0b\n\x03\x44OB\x18\t \x01(\x05\x12\x10\n\x08\x43HECKSUM\x18\n \x01(\t\x12\x12\n\nUPDATED_AT\x18\x0b \x02(\t')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'table_schema_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _TABLE_SCHEMA_PB2._serialized_start=23
  _TABLE_SCHEMA_PB2._serialized_end=228
# @@protoc_insertion_point(module_scope)



#
#   Created by Ryan McDermott
#   Created on 3/2/2022
#

import google.auth

from google.cloud import storage
from google.cloud import bigquery
from google.cloud import bigquery_storage_v1
from google.cloud import datastore
from google.cloud import pubsub_v1
from google.cloud import scheduler_v1
from google.cloud import tasks_v2
from google.cloud import secretmanager

from google.oauth2 import service_account
from google.cloud.compute_v1.services import instances
from google.cloud.compute_v1.services import zone_operations

class ApiClients:
    ''' Data Structure that hold all google API clients '''
    
    credentials                = ''
    scopes                          = ['https://www.googleapis.com/auth/cloud-platform']
    cloud_storage_client            = None
    big_query_client                = None
    secret_manager_client           = None
    big_query_storage_v1_client     = None
    datastore_client                = None
    pubsub_v1_client                = None
    cloud_tasks_client              = None
    cloud_scheduler_client          = None
    compute_engine_instance_client  = None
    compute_engine_operation_client = None
    account_credentials             = None
    account_googleapiclient         = None

    def __init__(self, credentials_file=None):
        '''
            Initializes all google api clients used in the data lake tools
            Args:
                credentials (string): The file path to the service acount key, or None
                    to use the default credentials
            Returns:
                None
        '''
        
        self.credentials = credentials_file
        
        if credentials_file is None:
            self.cloud_storage_client               = storage.Client()
            self.big_query_client                   = bigquery.Client()
            self.big_query_storage_v1_client        = bigquery_storage_v1.BigQueryWriteClient()
            self.secret_manager_client              = secretmanager.SecretManagerServiceClient()
            self.datastore_client                   = datastore.Client()
            self.pubsub_v1_client                   = pubsub_v1.PublisherClient()
            self.cloud_scheduler_client             = scheduler_v1.CloudSchedulerClient()
            self.compute_engine_instance_client     = instances.InstancesClient()
            self.compute_engine_operation_client    = zone_operations.ZoneOperationsClient()
            self.account_credentials, project       = google.auth.default()
            self.cloud_tasks_client                 = tasks_v2.CloudTasksClient()
            
        else:
            self.cloud_storage_client               = storage.Client.\
                from_service_account_json(credentials_file)
            self.big_query_client                   = bigquery.Client.\
                from_service_account_json(credentials_file)
            self.big_query_storage_v1_client        = bigquery_storage_v1.BigQueryWriteClient.\
                from_service_account_json(credentials_file)
            self.secret_manager_client              = secretmanager.SecretManagerServiceClient.\
                from_service_account_json(credentials_file)
            self.datastore_client                   = datastore.Client.\
                from_service_account_json(credentials_file)
            self.pubsub_v1_client                   = pubsub_v1.PublisherClient.\
                from_service_account_json(credentials_file)
            self.cloud_scheduler_client             = scheduler_v1.CloudSchedulerClient.\
                from_service_account_json(credentials_file)
            self.compute_engine_instance_client    = instances.InstancesClient.\
                from_service_account_json(credentials_file)
            self.compute_engine_operation_client    = zone_operations.ZoneOperationsClient.\
                from_service_account_json(credentials_file)
            self.account_credentials                = service_account.Credentials.\
                from_service_account_file(filename=credentials_file, scopes=self.scopes)
            self.cloud_tasks_client                 = tasks_v2.CloudTasksClient.\
                from_service_account_json(credentials_file)


    






#
#   Created by Ryan McDermott
#   Created on 2/25/2022
#

from indxdatalaketools import Helpers

class Standardizer():
    ''' Class that standardizes the file metadata dictionary '''
    arguments = {}


    def __init__(self, arguments):
        self.arguments = arguments

    
    def standardize(self, file_metadata):
        '''
            Standardizes the file metadata passed by the client, This includes adding the 
            modality, patient id, client id and file path if they do not exist
            Args:
                file_metadata (dict): the file metadata
            Returns:
                dict: The updated/standardized file metadata
        '''
        json_structure  = Helpers.determine_json_structure(file_metadata)
        file_metadata   = self.__populate_missing_fields(json_structure)
        file_metadata   = self.__set_missing_fields_to_unknown(file_metadata)

        return file_metadata
    

    def __populate_missing_fields(self, file_metadata):
        '''
            Populates the required metadata fields if they do not exist. These fields are
             modality, patient id, client id and file path
             Args:
                file_metadata (dict): The file metadata
            Returns:
                dict: updated file metadata
        '''
        if 'CLIENT_ID' not in file_metadata:
            file_metadata['CLIENT_ID'] = str(self.arguments['client_id'])

        if 'MODALITY' not in file_metadata:
            file_metadata['MODALITY'] = str(self.arguments['modality'])

        if 'PATIENT_ID' not in file_metadata:
            if self.arguments['mrn'] == 'UNKNOWN':
                file_metadata['PATIENT_ID'] = 'UNKNOWN'
            else:
                file_metadata['PATIENT_ID'] = Helpers.patient_hash(self.arguments['client_id'],
                [self.arguments['mrn']])[0]

        if 'FILE_PATH' not in file_metadata:
            file_metadata['FILE_PATH'] = str(file_metadata['MODALITY']) + \
            '/' + str(file_metadata['PATIENT_ID'])

        return file_metadata


    def __set_missing_fields_to_unknown(self, file_metadata):
        '''
            Sets any non-required missing fields to UNKOWN
            Args:
                file_metadata (dict): dictioary containing file metadata
            Retruns:
                dict: Updated file metadata
        '''
        optional_fields = ['DATE_OF_SERVICE', 'TIME_OF_SERVICE' ,'LOCATION_OF_SERVICE', 'SOURCE', 'ORIGINATOR']
        for field in optional_fields:
            if field not in file_metadata:
                file_metadata[field] = 'UNKNOWN'

        return file_metadata



#
#   Created By Ryan McDermott
#   Created On 5/13/2022
#
#   Top level file that hold all the logic for calling the executables. This
#   includes argument validations and implementing the sub modules for each 
#   command

import sys
import click
from indxdatalaketools import __version__
from indxdatalaketools.ClientTools import ClientOps

@click.group()
@click.version_option(__version__)
def main():
    """ ALL CLIENT OPERATIONS """


#####
#   File Commands
#####
@main.group()
def files():
    """Commands for file operations"""

@files.command('upload')
@click.option('--service_account', default=None, help='The path to the service account key json file')
@click.argument('client_id')
@click.argument('modality')
@click.argument('mrn')
@click.argument('metadata')
@click.argument('file_path')
def upload_file(client_id, modality, mrn, metadata, file_path, service_account):
    """\b\nupload a file
    
    Arguments: 
        
        CLIENT_ID:      The specific UUID4 for a client.
        MODALITY:       The files MODALITY. acceptable parameters are:
            HVF
            OCT_RNFL
            OCT_RETINA
            AVS
        MRN:            The Patients medical record number
        METADATA:       The json string or json file of the file you wish to upload.
                        The structure is:
            CLIENT_ID (REQUIRED):   String, UUID of client / health care agency
            PATIENT_ID:             String, UUID of patient
            MODALITY:               String, The modality of the file
            FILE_PATH:              String, The path in the datalake where the file is stored
            DATE_OF_SERVICE:        String, The date of service in
            TIME_OF_SERVICE:        String, The time of service in
            LOCATION_OF_SERVICE:    String, Where the file was created
            SOURCE:                 String, The Source of the file
            ORIGINATOR:             String, The user that uploaded the file to the data lake
        FILE_PATH:      The path to the file you wish to upload
        
    """
    file_client     = ClientOps.Client(client_id, credentials_file_path=service_account)
    result          = file_client.upload_file(modality, mrn, metadata, file_path)
    sys.exit(int(not result))


#####
#   PATIENTS Table Commands
#####
@main.group()
def patients_table():
    """Commands for PATIENTS table operations"""


@patients_table.command('upload')
@click.option('--service_account', default=None, help='The path to the service account key json file')
@click.argument('client_id')
@click.argument('patient_data')
def upload_patient(client_id, patient_data, service_account):
    """\b\nuploads a single, or multiple patients to the PATIENTS table
    
    Arguments: 
        
        CLIENT_ID:      The specific UUID4 for a client.
        PATIENT_DATA:   The json string, or json file representing a single, or 
                        multiple patients. The json structure is
            PATIENT_ID:             String, UUID of patient 
            CLIENT_ID (REQUIRED):   String, UUID of client / health care agency 
            MRN (REQUIRED):         String, patient’s medical record number per the client’s EMR system 
            LASTNAME:               String, patient’s last name 
            FIRSTNAME:              String, patient’s first name 
            MIDDLENAME:             String, patient’s middle name 
            DOB:                    Date, patient’s date of birth 
            SEX:                    String, patient’s administrative gender as per the HL7 code 
            RACE:                   String, patient’s race as per the HL7 code 
    """
    patients_table_client   = ClientOps.Client(client_id, service_account)
    result                  = patients_table_client.upload_patient(patient_data)
    sys.exit(int(not result))


if __name__ == '__main__':
    main()


import os
import setuptools
import site
import sys

site.ENABLE_USER_SITE = "--user" in sys.argv[1:]

dist = os.environ.get("INDX_DATALAKE_TOOLS_INSTALL", None)
print(dist)

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


if dist == 'admin':
    setuptools.setup(
        
        name="indxdatalaketools",
        
        setuptools_git_versioning={
            "enabled": True,
        },
        license = "Creative Commons Attribution-NonCommercial-NoDerivatives 4.0",
        license_files = ('LICEN[CS]E*', 'COPYING*', 'NOTICE*', 'AUTHORS*'),
        setup_requires=["setuptools-git-versioning"],
        data_files = [('man/man1', ['man/DataLakeTools.1']),],
        author="Ryan McDermott",
        author_email="rmcdermott@intelligentdx.com",
        description="Package that allows creation and manipulation of data lakes",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://github.com/IntelligentDX-LLC/data-lake-tools",
        project_urls={
            "Bug Tracker": "https://github.com/IntelligentDX-LLC/data-lake-tools/issues",
        },
        classifiers=[
            "Programming Language :: Python :: 3",
            "Operating System :: OS Independent",
        ],
        entry_points={'console_scripts': [
            'IndxAdminDataLakeTools = indxdatalaketools.AdminCommands.DataLakeToolsAdmin:main'
        ]},
        install_requires=[
        'google-cloud-datastore==2.5.1',
        'google-cloud-storage==2.3.0',
        'google-cloud-bigquery==3.1.0',
        'google-cloud-bigquery-storage==2.13.1',
        'google-auth==1.35.0',
        'google-cloud-pubsub==2.12.1',
        'google-cloud-scheduler==2.6.3',
        'google-cloud-secret-manager==2.10.0',
        'google-cloud-compute==0.7.0',
        'importlib-metadata==4.11.3',
        'protobuf==3.20.1',
        'click==8.1.3',
        'opencv-python==4.5.5.64',
        'google-api-python-client==1.9.3',
        'google-cloud-tasks==2.7.1',
        'requests==2.27.1',
        'numpy==1.22.3',
        'google-crc32c==1.3.0'
        ],
        packages=setuptools.find_packages(exclude=['**test**', 
        '*GoogleCloudResources*',]),
        python_requires=">=3.8",
        include_package_data=True,
        package_data={'': ['*ComputeEngineScripts/*']},
    )
else:
    setuptools.setup(
            
        name="indxdatalaketools",
        
        setuptools_git_versioning={
            "enabled": True,
            "template": "{tag}",
        },
        setup_requires=["setuptools-git-versioning"],
        license = "Creative Commons Attribution-NonCommercial-NoDerivatives 4.0",
        license_files = ('LICEN[CS]E*', 'COPYING*', 'NOTICE*', 'AUTHORS*'),
        data_files = [('man/man1', ['man/DataLakeTools.1']),],
        author="Ryan McDermott",
        author_email="rmcdermott@intelligentdx.com",
        description="Package that allows the upload of files to a datalake",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://github.com/IntelligentDX-LLC/data-lake-tools",
        project_urls={
            "Bug Tracker": "https://github.com/IntelligentDX-LLC/data-lake-tools/issues",
        },
        classifiers=[
            "Programming Language :: Python :: 3",
            "Operating System :: OS Independent",
        ],
        entry_points={'console_scripts': [
            'IndxClientDataLakeTools = indxdatalaketools.ClientCommands.DataLakeToolsClient:main'
        ]},
        install_requires=[
        'google-cloud-datastore==2.5.1',
        'google-cloud-storage==2.3.0',
        'google-cloud-bigquery==3.1.0',
        'google-cloud-bigquery-storage==2.13.1',
        'google-auth==1.35.0',
        'google-cloud-pubsub==2.12.1',
        'google-cloud-scheduler==2.6.3',
        'google-cloud-secret-manager==2.10.0',
        'google-cloud-compute==0.7.0',
        'importlib-metadata==4.11.3',
        'protobuf==3.20.1',
        'click==8.1.3',
        'opencv-python==4.5.5.64',
        'google-api-python-client==1.9.3',
        'google-cloud-tasks==2.7.1',
        'requests==2.27.1',
        'numpy==1.22.3',
        'google-crc32c==1.3.0',
        ],
        packages=setuptools.find_packages(
            include=['*ArgumentValidation*', 
            '*DataStandardizer*', 
            '*GoogleClientWrappers*', 
            '*ClientTools*',
            '*ClientCommands*',
            'indxdatalaketools',
            '*data*'], 
            exclude=['**test**', '**pycache**',
            '*GoogleCloudResources*',
            '*ComputeEngineCreator*',
            '*ComputeEngineScripts*',
            '*DataLakeFileOperations*',
            '*DataLakeOperations*',
            '*AdminTools*',]),
        python_requires=">=3.8",
        include_package_data=True,
        package_data={'': []},
    )
